/**@name LIBRARY: unLHCLoggingDBSql.ctl

@author: Josef Hofer (EN-ICE-SCD)

Creation Date: 04/09/2014

Purpose:
This library contains database related functions for the unLHCLoggingDB component.

Usage: Public

PVSS manager usage: CTRL, UI

Constraints:
	. PVSS version: 3.11
	. operating system: Linux.

Copyright CERN

*/

//@{


// Version of this library.
const string c_unLHCLoggingDBSql_LibVersion = "1.0.5";


// Database connection
global dbConnection g_unLHCLoggingDBSql_dbConnection;


/**
* Connect to db, if the connectionName exists, then use the existing. If connexction fails return FALSE.
*
* @param sDbName   input, database name
* @param sDbUser   input, database user
* @param sDbPass   input, database password (encrypted)
*
* @return          output, true: success, false: error
*/
public bool unLHCLoggingDBSql_openDbConnection(string sDbName, string sDbUser, string sDbPass)
{
    // Specify connection name 
    string connectionName = "unLHCLoggingDB_"+sDbName+"_"+sDbUser;

    string errTxt;
    int rc = fwDbGetConnection(connectionName, g_unLHCLoggingDBSql_dbConnection);
    if (rc == -1) {
        fwDbCheckError(errTxt, g_unLHCLoggingDBSql_dbConnection);
        DebugTN("ERROR Getting the connection", errTxt);
        return false;
    } else if (rc == 1) { // connection with such name does not exist yet. Create it
        string connectString = "Driver=QOCI;Database=" + sDbName + ";User=" + sDbUser + ";ENC_PWD=" + sDbPass;
        fwDbOpenConnection(connectString, g_unLHCLoggingDBSql_dbConnection, connectionName);
        if (fwDbCheckError(errTxt, g_unLHCLoggingDBSql_dbConnection)) {DebugTN("ERROR WHILE OPENING DATABASE", errTxt); return false;};
    }
    //Add a debug info for existed connection
    mixed retVal; // note! the argument to fwDbOption must be of type mixed!
    int rc = fwDbOption("Inspect", g_unLHCLoggingDBSql_dbConnection, retVal);
    DebugTN(rc, retVal);
    return true;
}


/**
* Start execution of a db command
*
* @param cmd     input, command
* @param sql     input, sql query text
*
* @return true:success, false:error
*/
public bool unLHCLoggingDBSql_startCommand(dbCommand &cmd, string &sql)
{
    string errTxt;

    fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sql, cmd);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;}; 

    return true;    
}


/**
* Execute a db command with paramter mapping and get result
*
* @param cmd     input, command
* @param map     input, parameter mapping
* @param result  output, result as dyn_dyn_mixed
*
* @return true:success, false:error
*/
public bool unLHCLoggingDBSql_executeCommand(dbCommand &cmd, mapping &map, dyn_dyn_mixed &result, bool columnWise=false)
{    
    string errTxt;   
       
    fwDbBindParams(cmd, map);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};

    fwDbExecuteCommand(cmd);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};
    
    fwDbGetData(cmd, result, columnWise);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};  

    return true;    
}


/**
* Finish execution of a db command
*
* @param cmd     input, command
* @param sql     input, sql query text
*
* @return true:success, false:error
*/
public bool unLHCLoggingDBSql_finishCommand(dbCommand &cmd)
{
    string errTxt;

    fwDbFinishCommand(cmd);
    if (fwDbCheckError(errTxt)) {DebugTN("DB ERROR", errTxt); return false;};

    return true;    
}


/**
* Execute sql query with parameter mapping
*
* @param sqlQuery   input, SQL query string
* @param params     input, parameter mapping
* @param dsExInfo   output, Exception information
* @param columnWise input, false: get result row wise, true: get result column wise
* @param blocking   input, execute blocking queries on default, but we can also switch to asynchronous mode
*
* @return data      output, result of the query
*/
public dyn_dyn_mixed unLHCLoggingDBSql_queryDataWithParameterMapping(string sqlQuery, mapping params, dyn_string &dsExInfo, bool columnWise=false, bool blocking=true)
{  
    string errTxt;
    dbCommand cmd;
    dyn_dyn_mixed data; 
    
    if(blocking) {  
  
        mixed retVal; 
        int rc = fwDbOption("SetCommandTimeout", -1, retVal);      

        fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sqlQuery, cmd);
        if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};

        if (mappinglen(params)) {
            fwDbBindParams(cmd, params);
            if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};
        }

        fwDbExecuteCommand(cmd);
        if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};

        fwDbGetData(cmd, data, columnWise);
        if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};

        fwDbFinishCommand(cmd);
        if (fwDbCheckError(errTxt)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};

        return data;   
        
    } else { // asynchronous    

        mixed retVal; 
        int rc = fwDbOption("SetCommandTimeout", 0, retVal);  

        fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sqlQuery, cmd);
        if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};        

        if (mappinglen(params)) {
            fwDbBindParams(cmd, params);
            if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};
        }

        fwDbExecuteCommand(cmd);
        if (fwDbCheckError(errTxt, cmd)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};  
                   
        time tStart = getCurrentTime();
        do {                    
            delay(0,250);
            time tNow = getCurrentTime();
            if(tNow-tStart >= 30) {
                fwDbBreak(g_unLHCLoggingDBSql_dbConnection); // abort command execution   
                fwException_raise(dsExInfo, "DB ERROR", "timeout executing query: " + sqlQuery, ""); 
                return data;
            }       
        } while ( fwDbIsConnectionBusy(g_unLHCLoggingDBSql_dbConnection) );  
        
        // Get data in blocks of 1000 records and add some waitstates to don't block other threads too long        
        while(1) {
            dyn_dyn_mixed tempData; 
            int ret = fwDbGetData(cmd, tempData, false, 1000, makeDynInt(), 30);  
            if(dynlen(tempData)>0) { 
                // fwDbGetData is inserting an empty dyn_mixed into tempData when no data is retrieved
                // when tempData is appended to data, we get dyn_dyn_mixed with dynlen==1, 
                // because it contains a dyn_mixed with dynlen==0
                // this does not make sense at all and will cause problems
                // workaround: only append tempData to data, if tempData is not empty
                dynAppend(data, tempData);
            } 
            if(ret == 1) { delay(0,100); } //	success and there are still rows to extract
            else if(ret == 0) { break; }   // success, and no rows left to extract 
            else { fwException_raise(dsExInfo, "DB ERROR", "timeout getting data for query: " + sqlQuery, ""); return data; }            
        }  
        // Turn the array to get data column wise        
        if(columnWise) { 
            dyn_dyn_anytype ddaData = data;
            dynDynTurn(ddaData); // workaround because dynDynTurn does not work for dyn_dyn_mixed
            data = ddaData;
        }        

        fwDbFinishCommand(cmd);
        if (fwDbCheckError(errTxt)) {fwException_raise(dsExInfo, "DB ERROR", errTxt, ""); return data;};            
    }
    
    return data;
}


/**
* Execute sql query
*
* @param sqlQuery   input, SQL query string
* @param dsExInfo   output, Exception information
* @param columnWise input, false: get result row wise, true: get result column wise
*
* @return data      output, result of the query
*/
public dyn_dyn_mixed unLHCLoggingDBSql_queryData(string sqlQuery, dyn_string &dsExInfo, bool columnWise=false, bool blocking=true)
{
    mapping params;
    return unLHCLoggingDBSql_queryDataWithParameterMapping(sqlQuery, params, dsExInfo, columnWise, blocking);
}


/**
* Execute sql command with bound data
*
* Using this function we will gain speed by sending a sql command just once (e.g. an insert into statement)
* but executing it many times with different input data (provided through the dyn_dyn_mixed input-parameter)
* unLHCLoggingDBSql_executeWithBoundData("INSERT INTO myTable(colId,colName,colUnit) VALUES(:data1,:data2,:data3)");
*
* @param sql       input, SQL command string
* @param data      input, data to be inserted
*
* @return          output, true: success, false: error
*/
public bool unLHCLoggingDBSql_executeWithBoundData(string sql, dyn_dyn_mixed &data)
{
    // Insert data records into table GTT_STAGE_VARIABLES
    dbCommand cmd;
    string errTxt;

    fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sql, cmd);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};

    fwDbBindAndExecute(cmd, data);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};

    fwDbFinishCommand(cmd);
    if (fwDbCheckError(errTxt)) {DebugTN("DB ERROR", errTxt); return false;};

    return true;
}


/**
* Execute sql command with parameter mapping
*
* @param sql       input, SQL command string
* @param params    input, Parameter mapping like: param["user"]=sUserName;
*
* @return          output, true: success, false: error
*/
public bool unLHCLoggingDBSql_executeWithParameterMapping(string sql, mapping params)
{
    string errTxt;
    dbCommand cmd;

    fwDbStartCommand(g_unLHCLoggingDBSql_dbConnection, sql, cmd);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};

    if (mappinglen(params)) {
        fwDbBindParams(cmd, params);
        if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};
    }

    fwDbExecuteCommand(cmd);
    if (fwDbCheckError(errTxt, cmd)) {DebugTN("DB ERROR", errTxt); return false;};

    fwDbFinishCommand(cmd);
    if (fwDbCheckError(errTxt)) {DebugTN("DB ERROR", errTxt); return false;};

    return true;
}


/**
* Execute sql command 
*
* @param sql       input, SQL command string
*
* @return          output, true: success, false: error
*/
public bool unLHCLoggingDBSql_execute(string sql)
{
    mapping params;
    return unLHCLoggingDBSql_executeWithParameterMapping(sql, params);
}


/**
* commit DB operation
*
* @param   dsExInfo      output: Exception info
*/
public void unLHCLoggingDBSql_commit(dyn_string &dsExInfo)
{
    if (unLHCLoggingDBSql_execute("begin commit; end;") == false) {
        fwException_raise(dsExInfo, "DB ERROR", "on commit", "");
    }
}


/**
* Get all DB schemas, which have an ELEMENTS table
*
* @param   dsExInfo      output: Exception info
* @return  dsDbSchemas   output: List of DB schemas
*/
public dyn_string unLHCLoggingDBSql_getDbSchemas(dyn_string &dsExInfo)
{
    string sql = "SELECT owner " +
                 " FROM all_tables " +
                 "WHERE table_name='ELEMENTS' " +
                 " ORDER BY owner ASC";   
    
    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo);

    if (dynlen(dsExInfo) > 0) {
        return makeDynString("");
    }

    if (dynlen(sqlResult) <= 0) {
        fwException_raise(dsExInfo, "DB ERROR", "No database schemas found", "");
        return makeDynString("");
    }

    dyn_string dsDbSchemas = getDynMixed(sqlResult, 1); //Get column 1
    dynAppend(dsDbSchemas, "*");  // add star to select all tables
    return dsDbSchemas;
}


/**
* Get all RDB Sources of a specific DB schema
*
* @param   sDbSchema    input,  Database schema
* @param   dsExInfo     output: Exception info
*
* @return  sqlResult    output, list of distinct rdb sources
*
*
*/
public dyn_string unLHCLoggingDBSql_getRdbSource(string sDbSchema, dyn_string &dsExInfo)
{
    string sql = "select distinct RDB_SOURCE "
                 "from RDB2LHC.META_VARIABLES "
                 "where RDB_SOURCE like '" + sDbSchema + ".%'";
    
    return unLHCLoggingDBSql_queryData(sql, dsExInfo);
}


/**
* Get all variables of a specified db schema that are registered for logging and retrieve detail informations
*
* @param   sDbSchema    input,  Database schema
* @param   bUTC         input,  true=show timestamps as UTC, false=show timestamps as localtime
* @param   dsExInfo     output: Exception info
*
* @return  sqlResult    output, Query result (column wise)
*
* line  1: Variable ID
* line  2: Variable Name
* line  3: Dpe NAME
* line  4: Dpe ALIAS
* line  5: Data category
* line  6: User name
* line  7: Logging required
* line  8: Registered for logging
* line  9: Rdb source
* line 10: Hierarchy
* line 11: Transfer job ID
* line 12: Timestamp last job run
* line 13: Timestamp last logged
* line 14: No records last logged
* line 15: description
*
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getLogVarInfo(string sDbSchema, dyn_string &dsExInfo, bool bUTC = false, bool bGetHierarchies = true)
{
    dyn_dyn_mixed ddmResult;
    
    // Prequery rdb sources 
    dyn_string dsRdbSource = unLHCLoggingDBSql_getRdbSource(sDbSchema,dsExInfo);
    if (dynlen(dsExInfo)>0) return ddmResult; // return empty result
    
    int dsRdbSourceLen = dynlen(dsRdbSource);
    for(int i=1; i<=dsRdbSourceLen; i++) { 
        string sql = "SELECT " + 
                     "  TO_CHAR(mv.variable_id), " +  // get variable_id as a char because it has too many digits to display it as a number
                     "  mv.variable_name, " + // get variable name
                     "  el.element_name, " +  // get datapoint element name
                     "  el.alias, " + // get datapoint element alias
                     "  mvc.data_category_name || '/' || mvc.data_category_sub_name, " + // get data_category_name/data_category_sub_name
                     "  mu.real_name, " +  // get username
                     "  REPLACE(REPLACE(mv.logging_required,'Y','YES'),'N','NO'), " + // get logging_required as "YES" or "NO"
                     "  REPLACE(REPLACE(mv.registered_for_logging,'Y','YES'),'N','NO'), " + // get registered_for_logging as "YES" or "NO"
                     "  mv.rdb_source, " + // get rdb_source = dbschema + "." + archive_group + "HISTORY"
                      ( bGetHierarchies ? " REPLACE(mh.node_path,'->','/'), " : " TO_CHAR(''), " ) + // get meta hierarchies replace -> by /
                     "  mvc.data_transfer_group_id, " + // get transfer job id
                     "  TO_CHAR(" + ( bUTC ? "lvs.created_utc_stamp" : "FROM_TZ(lvs.created_utc_stamp, 'UTC') AT LOCAL" ) + ",'YYYY/MM/DD HH24:MI:SS'), " + // get timestamp as utc or localtime
                     "  TO_CHAR(" + ( bUTC ? "lvs.last_logged_utc_stamp" : "FROM_TZ(lvs.last_logged_utc_stamp, 'UTC') AT LOCAL" ) + ",'YYYY/MM/DD HH24:MI:SS'), " + // get timestamp as utc or localtime
                     "  lvs.no_recs_logged, " +  // get number of records logged
                     "  mv.description " // get variable description
                     "FROM rdb2lhc.meta_variables mv "
                     "  LEFT OUTER JOIN " + sDbSchema + ".elements el ON mv.element_id = el.element_id " + // to get dpe names
                     "  LEFT OUTER JOIN rdb2lhc.meta_variable_categories mvc ON mv.data_category_id = mvc.data_category_id " + // to get data_category names
                     "  LEFT OUTER JOIN rdb2lhc.meta_users mu ON mv.owner = mu.user_name " + // to get usernames
                      ( bGetHierarchies ? " LEFT OUTER JOIN rdb2lhc.meta_hierarchy_variables mhv ON mv.variable_id = mhv.variable_id " : "") + // to get hierarchies
                      ( bGetHierarchies ? " LEFT OUTER JOIN rdb2lhc.meta_hierarchies mh ON mhv.hierarchy_id = mh.hierarchy_id " : "") +  // to get hierarchies
                     "  LEFT OUTER JOIN rdb2lhc.log_var_status_max lvsm ON mvc.data_transfer_group_id = lvsm.data_transfer_group_id " + // to get timestamps
                     "  LEFT OUTER JOIN rdb2lhc.log_variable_status lvs ON mv.variable_id = lvs.variable_id " + // to get timestamps of last transfer
                     "                                                AND mvc.data_transfer_group_id = lvs.data_transfer_group_id " + // to get timestamps of last transfer
                     "                                                AND lvsm.created_utc_stamp = lvs.created_utc_stamp " + // to get timestamps of last transfer
                     "WHERE mv.rdb_source = '" + dsRdbSource[i] + "' AND mv.obsolete_since_utc IS NULL"; // filter only for specified dbSchema and not obsolete

        dyn_dyn_mixed ddmTempResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, true, false); 
        if (dynlen(dsExInfo)>0) return ddmResult; // return empty result
           
        int  ddmTempResultLen = dynlen(ddmTempResult);    
        for(int i=1; i<=ddmTempResultLen; i++) {
            dynAppend(ddmResult[i], ddmTempResult[i]); 
        }
    }
    
    return ddmResult;
}

/**
* Get all variables of a specified db schema that are registered for logging
*
* @param   sDbSchema    input,  Database schema
* @param   dsExInfo     output: Exception info
*
* @return  sqlResult    output, Query result (column wise)
*
* line  1: Variable Name
* line  2: Data category
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getLogVariables(string sDbSchema, dyn_string &dsExInfo)
{
    dyn_dyn_mixed ddmResult;
    
    // Prequery rdb sources 
    dyn_string dsRdbSource = unLHCLoggingDBSql_getRdbSource(sDbSchema,dsExInfo);
    if (dynlen(dsExInfo)>0) return ddmResult; // return empty result
    
    int dsRdbSourceLen = dynlen(dsRdbSource);
    for(int i=1; i<=dsRdbSourceLen; i++) { 
        string sql = "SELECT " + 
                     "  mv.variable_name, " + // get variable name
                     "  mvc.data_category_name || '/' || mvc.data_category_sub_name " + // get data_category_name/data_category_sub_name
                     "FROM rdb2lhc.meta_variables mv "
                     "  LEFT OUTER JOIN " + sDbSchema + ".elements el ON mv.element_id = el.element_id " + // to get dpe names
                     "  LEFT OUTER JOIN rdb2lhc.meta_variable_categories mvc ON mv.data_category_id = mvc.data_category_id " + // to get data_category names
                     "WHERE mv.rdb_source = '" + dsRdbSource[i] + "' "
                     "  AND mv.obsolete_since_utc IS NULL"; // filter only for specified dbSchema and not obsolete

        dyn_dyn_mixed ddmTempResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, true, false); 
        if (dynlen(dsExInfo)>0) return ddmResult; // return empty result
           
        int  ddmTempResultLen = dynlen(ddmTempResult);    
        for(int i=1; i<=ddmTempResultLen; i++) {
            dynAppend(ddmResult[i], ddmTempResult[i]); 
        }
    }
    
    return ddmResult;
}


/**
* Get number of registered elements of a specified database schema
*
* @param   sDbSchema    input,  Database schema
* @param   dsExInfo     output  Exception info
*
* @return  sqlResult    output, No of elements
*
*/
public unsigned unLHCLoggingDBSql_getLogVariablesCount(string sDbSchema, dyn_string &dsExInfo)
{
    string sql = "SELECT count(1) " +
                 "FROM rdb2lhc.meta_variables mv " +
                 "WHERE mv.rdb_source LIKE '" + sDbSchema + ".%' " + 
                 "  AND mv.obsolete_since_utc IS NULL";

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo);

    if (dynlen(sqlResult) > 0) {
        return sqlResult[1][1];
    }

    return 0;
}


/**
* Get details of a particular element
*
* Use this function only to query single elements because it is not
* very fast, which is caused by the prequery of the dbSchema name
*
* @param   sVariableId   input,  Variable ID of the selected element
* @param   dsExInfo      output, ExceptionInfo

* @return  sqlResult      output, Result of the query (row wise)
*
* column 1: Variable name
* column 2: Datapoint name
* column 3: Data category
* column 4: User name
* column 5: Logging requred
* column 6: Registered for logging
* column 7: Rdb source
* column 8: Hierarchy
*
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getElementDetails(string sVariableId, dyn_string &dsExInfo)
{
    mapping params;
    params["VarId"] = sVariableId;

    string sql1 = "SELECT SUBSTR(rdb_source, 1, INSTR4(rdb_source, '.')-1) "
                  "FROM rdb2lhc.meta_variables "
                  "WHERE variable_id = :VarId";

    // Get name of db schema
    dyn_dyn_mixed sqlPreResult = unLHCLoggingDBSql_queryDataWithParameterMapping(sql1, params, dsExInfo);

    dyn_dyn_mixed sqlResult;
    if (dynlen(dsExInfo) > 0) {return sqlResult;}
    if (dynlen(sqlPreResult) <= 0) {fwException_raise(dsExInfo, "DB ERROR", "No database schema found", ""); return sqlResult; };

    string sDbSchema = sqlPreResult[1][1];

    string sql2 = "SELECT mv.variable_name, " +
                  "el.element_name, " +
                  "mvc.data_category_name || '/' || mvc.data_category_sub_name, " +
                  "mu.real_name, " +
                  "REPLACE(REPLACE(mv.logging_required,'Y','1'),'N','0'), " +
                  "REPLACE(REPLACE(mv.registered_for_logging,'Y','1'),'N','0'), " +
                  "mv.rdb_source, " +
                  "REPLACE(mh.node_path,'->','/'), " +
                  "mv.description "
                  "FROM rdb2lhc.meta_variables mv "
                  "LEFT OUTER JOIN " + sDbSchema + ".elements el ON mv.element_id = el.element_id " +
                  "LEFT OUTER JOIN rdb2lhc.meta_variable_categories mvc ON mv.data_category_id = mvc.data_category_id " +
                  "LEFT OUTER JOIN rdb2lhc.meta_users mu ON mv.owner = mu.user_name " +
                  "LEFT OUTER JOIN rdb2lhc.meta_hierarchy_variables mhv ON mv.variable_id = mhv.variable_id " +
                  "LEFT OUTER JOIN rdb2lhc.meta_hierarchies mh ON mhv.hierarchy_id = mh.hierarchy_id " +
                  "WHERE mv.variable_id = :VarId";

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql2, params, dsExInfo); // Get data line wise
}


/**
* Get the transfer history of a specific element (MDB -> LDB)
*
* @param   sVariableId   input,  Variable ID of the selected element
* @param   tStart        input,  Query start time (UTC only)
* @param   tEnd          input,  Query end time (UTC only)
* @param   bUTC          input,  Display timestamps of query result in UTC or LOCAL
* @param   dsExInfo      output, ExceptionInfo

* @return  sqlResult      output, Result of the query (column wise)
*
* line 1: Time created
* line 2: Time first checked
* line 3: Time last checked
* line 4: No of records checked
* line 5: Time last logged
* line 6: No of records logged
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getTransferHistory(string sVariableId, time tStart, time tEnd, bool bUTC = false, dyn_string &dsExInfo)
{
    string sql = "select to_char(" + ( bUTC ? "CREATED_UTC_STAMP" : "from_tz(CREATED_UTC_STAMP, 'UTC') at local" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
                 " to_char(" + ( bUTC ? "FIRST_CHECKED_UTC_STAMP" : "from_tz(FIRST_CHECKED_UTC_STAMP, 'UTC') at local" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
                 " to_char(" + ( bUTC ? "LAST_CHECKED_UTC_STAMP" : "from_tz(LAST_CHECKED_UTC_STAMP, 'UTC') at local" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
                 " NO_RECS_CHECKED, " +
                 " to_char(" + ( bUTC ? "LAST_LOGGED_UTC_STAMP" : "from_tz(LAST_LOGGED_UTC_STAMP, 'UTC') at local" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
                 " NO_RECS_LOGGED " +
                 "from RDB2LHC.LOG_VARIABLE_STATUS " +
                 " where variable_id = :varId and CREATED_UTC_STAMP BETWEEN :t1 and :t2 " +
                 " order by created_utc_stamp desc";

    mapping params;
    params[":varId"] = sVariableId;
    params[":t1"] = tStart;
    params[":t2"] = tEnd;

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo, true, false); // TODO delete last parameter
}


/**
* Get value history of a specific meta variable
*
* @param   sVariableId  input,  VARIABLE_ID of the selected element
* @param   tStart       input,  Query start time (UTC only)
* @param   tEnd         input,  Query end time (UTC only)
* @param   bUTC         input,  Display timestamps of query result in UTC or LOCAL
* @param   dsExInfo     output, ExceptionInfo

* @return  sqlResult    output, Result of the query (column wise)
*
* line 1: Time archived to MDB
* line 2: Value
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getElementHistory(string sVariableId, time tStart, time tEnd, bool bUTC = false, dyn_string &dsExInfo)
{
    // Prequery for RDB_SOURCE and ELEMENT_ID
    mapping params1;
    params1["VarId"] = sVariableId;

    string sql1 = "SELECT rdb_source " +
                  "     , TO_CHAR(element_id) " +
                  "FROM rdb2lhc.meta_variables " +
                  "WHERE variable_id = :VarId" ;

    dyn_dyn_mixed sqlPreResult = unLHCLoggingDBSql_queryDataWithParameterMapping(sql1, params1, dsExInfo);

    dyn_dyn_mixed sqlResult;
    if (dynlen(dsExInfo) > 0) { return sqlResult; };
    if (dynlen(sqlPreResult) <= 0) {fwException_raise(dsExInfo, "DB ERROR", "Variable not found", ""); return sqlResult; };

    string sRdbSource = sqlPreResult[1][1];
    string sElementId = sqlPreResult[1][2];

    // Main query for history data
    mapping params2;
    params2[":elementId"] = sElementId;
    params2[":t1"] = tStart;
    params2[":t2"] = tEnd;

    string sql2 = "SELECT TO_CHAR(" + ( bUTC ? "ts" : "FROM_TZ(ts, 'UTC') AT LOCAL" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
                  " TO_NUMBER(value_number) " +
                  "FROM " + sRdbSource +
                  " WHERE element_id = :elementId AND ts BETWEEN :t1 AND :t2" +
                  " ORDER BY ts DESC";

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql2, params2, dsExInfo, true);
}


/**
* Get meta users from RDB2LHC.META_USERS
*
* @param dsUserNames    output, list of user names
* @param dsRealNames    output, list of real names
* @param dsExInfo       output, Exception information
*/
public void unLHCLoggingDBSql_getMetaUsers(dyn_string &dsUserNames, dyn_string &dsRealNames, dyn_string &dsExInfo)
{
    dynClear(dsUserNames);
    dynClear(dsRealNames);

    string sql = "SELECT user_name, real_name FROM rdb2lhc.meta_users";
    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, true);

    if (dynlen(dsExInfo) > 0) { return; };
    if (dynlen(sqlResult) != 2) { return; }; // there must be 2 lines

    dsUserNames = sqlResult[1];
    dsRealNames = sqlResult[2];
}


/**
* Check if user pass combination is found in the database
*
* @param sUser    output, username
* @param sPass    output, password
*
* @return         output, true: success, false: error
*/
public bool unLHCLoggingDBSql_checkUserPass(string sUser, string sPass)
{
    dyn_string dsExInfo;

    string sql = "SELECT user_name "
                 "  FROM rdb2lhc.meta_users "
                 " WHERE user_name = :username "
                 "   AND password = :password";

    mapping params;
    params[":username"] = sUser;
    params[":password"] = sPass;

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo);

    if (dynlen(dsExInfo) > 0) { return false; }
    if (dynlen(sqlResult) <= 0) { return false; }
    if (sqlResult[1][1] == sUser) { return true; }
    return false;
}


/**
* Get credentials for user SystemIntegrity
*
* @param sUser    output, username
* @param sPass    output, password
* @param dsExInfo output, Exception information
*/
public void unLHCLoggingDBSql_getSystemIntegrityCredentials(string &sUser, string &sPass, dyn_string &dsExInfo)
{
    string sql = "SELECT user_name, password "
                 "  FROM rdb2lhc.meta_users "
                 " WHERE user_name = 'SystemIntegrity'";

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo);

    if (dynlen(dsExInfo) > 0) { return; }
    if (dynlen(sqlResult) <= 0) {fwException_raise(dsExInfo, "DB ERROR", "Credentials not found", ""); return; };

    sUser = sqlResult[1][1];
    sPass = sqlResult[1][2];
}


/**
* Create a new data category
*
* @param sName            input, data category name
* @param sSubName         input, data category sub name
* @param sTransferGroupId input, transfer group id
*
* @return                 output, true: sucess, false:error
*/
public bool unLHCLoggingDBSql_createDataCategory(string sName, string sSubName, string sTransferGroupId)
{
    string sql = "declare "
                 " catId INTEGER; "
                 "begin "
                 " pvss_data_transfer_mgr.add_data_category(:catName, :catSubName, :catGroupId, catId); "
                 "end;";

    mapping params;
    params[":catName"]    = sName;
    params[":catSubName"] = sSubName;
    params[":catGroupId"] = sTransferGroupId;

    return unLHCLoggingDBSql_executeWithParameterMapping(sql, params);
}


/**
* Update transfer group of an existing data category
*
* @param sDataCatName     input, data category name
* @param sDataCatSubName  input, data category sub name
* @param sTransferGroupId input, transfer group id
*
* @return                 output, true: sucess, false:error
*/
public bool unLHCLoggingDBSql_updateTransferGroupId(string sDataCatName, string sDataCatSubName, string sTransferGroupId)
{
    string sql = "begin "
                 " pvss_data_transfer_mgr.update_data_category_trans_grp(:catName, :catSubName, :catGroupId); "
                 "end;";

    mapping params;
    params[":catName"]    = sDataCatName;
    params[":catSubName"] = sDataCatSubName;
    params[":catGroupId"] = sTransferGroupId;

    return unLHCLoggingDBSql_executeWithParameterMapping(sql, params);
}


/**
* Get data categories from RDB2LHC.META_VARIABLE_CATEGORIES
*
* @param   dsExInfo             output, Exception information
* @return  sqlResult            output, data categories
*
* line 1: data category id
* line 2: data category name
* line 3: data category sub name
* line 4: data category name / data category sub name
* line 5: data transfer group id
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getDataCategories(dyn_string &dsExInfo)
{
    string sql = "SELECT data_category_id, " +
                 "  data_category_name, " +
                 "  data_category_sub_name, " +
                 "  data_category_name || '/' || data_category_sub_name, " +
                 "  data_transfer_group_id " +
                 "FROM rdb2lhc.meta_variable_categories " +
                 "  ORDER BY data_category_name ASC";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, true);
}


/**
* Get available data categories for Specific Schema and User
*
* @param   sSchema              input, schema name
* @param   sUser                input, user name
* @param   dsExInfo             output, Exception information
*
* @return  sqlResult            output, data categories
*
* line 1: data category id
* line 2: data category name
* line 3: data category sub name
* line 4: data category name / data category sub name
* line 5: data transfer group id
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getDataCategoriesForSpecific(string sSchema, string sUser, dyn_string &dsExInfo)
{   
    string sql ="SELECT "
                "  mvc.data_category_id, "
                "  mvc.data_category_name, "
                "  mvc.data_category_sub_name, "
                "  mvc.data_category_name || '/' || mvc.data_category_sub_name, "
                "  mvc.data_transfer_group_id "
                "FROM "
                "  meta_category_mapping mcm "
                "JOIN rdb2lhc.meta_variable_categories mvc "
                "  ON mcm.data_category_id = mvc.data_category_id "
                "WHERE mcm.user_name = :userName AND mcm.schema_name = :schemaName "
                "  ORDER BY data_category_name ASC";

    mapping params;    
    params[":schemaName"] = sSchema;
    params[":userName"]   = sUser;

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo, true);
}


/**
* Get mapping between data category, schema and user
*
* @param   dsExInfo             output, Exception information
*
* @return  sqlResult            output, data cagegory mapping (column wise)
*
* column 1: data category name/subname
* column 2: schema name
* column 3: user name
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getDataCategoryMapping(dyn_string &dsExInfo)
{   
    string sql ="SELECT "
                "  mvc.data_category_name || '/' || mvc.data_category_sub_name cat, "
                "  mcm.schema_name, "
                "  mcm.user_name "
                "FROM "
                "  rdb2lhc.meta_category_mapping mcm "
                "JOIN "
                "  rdb2lhc.meta_variable_categories mvc "
                "ON "
                "  mcm.data_category_id = mvc.data_category_id "
                "ORDER BY "
                "  cat ";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, true);
}


/**
* Add mapping between data category, schema and user
*
* @param   sUser                input, user name
* @param   sSchema              input, schema name
* @param   sDataCategory        input, data categrory id as string
* @param   dsExInfo             output, Exception information
*/
public void unLHCLoggingDBSql_addDataCategoryMapping(string sUser, string sSchema, string sDataCategory, dyn_string &dsExInfo)
{  
    string sql = "INSERT INTO meta_category_mapping "
                 "  (user_name, schema_name, data_category_id) "
                 "VALUES "
                 "  (:userName, :schemaName, :dataCatId)";

    mapping params;   
    params[":userName"]  = sUser; 
    params[":schemaName"] = sSchema;      
    params[":dataCatId"]  = sDataCategory; 
    
    if (unLHCLoggingDBSql_executeWithParameterMapping(sql, params) == false) {
        fwException_raise(dsExInfo, "DB ERROR", "adding new data category mapping", "");
        return;
    }
    
    unLHCLoggingDBSql_commit(dsExInfo);
}


/**
* Update mapping between data category, schema and user
*
* @param   sNewUser             input, user name
* @param   sNewSchema           input, schema name
* @param   sNewDataCategory     input, data categrory id as string
* @param   sOldUser             input, user name
* @param   sOldSchema           input, schema name
* @param   sOldDataCategory     input, data categrory id as string
* @param   dsExInfo             output, Exception information
*/
public void unLHCLoggingDBSql_updateDataCategoryMapping(     
        string sNewUser, 
        string sNewSchema, 
        string sNewDataCategory,
        string sOldUser, 
        string sOldSchema, 
        string sOldDataCategory,    
        dyn_string &dsExInfo)
{  
    string sql = "UPDATE meta_category_mapping "
                 "   SET user_name=:newUserName, schema_name=:newSchemaName, data_category_id=:newDataCatId "
                 "WHERE user_name=:oldUserName "
                 "  AND schema_name=:oldSchemaName "
                 "  AND data_category_id=:oldDataCatId ";

    mapping params;   
    params[":newUserName"]  = sNewUser; 
    params[":newSchemaName"] = sNewSchema;      
    params[":newDataCatId"]  = sNewDataCategory; 
    params[":oldUserName"]  = sOldUser; 
    params[":oldSchemaName"] = sOldSchema;      
    params[":oldDataCatId"]  = sOldDataCategory; 
    
    if (unLHCLoggingDBSql_executeWithParameterMapping(sql, params) == false) {
        fwException_raise(dsExInfo, "DB ERROR", "updating data category mapping", "");
        return;
    }
    
    unLHCLoggingDBSql_commit(dsExInfo);
}


/**
* Remove mapping between data category, schema and user
*
* @param   sUser                input, user name
* @param   sSchema              input, schema name
* @param   sDataCategory        input, data categrory id as string
* @param   dsExInfo             output, Exception information
*/
public void unLHCLoggingDBSql_deleteDataCategoryMapping(string sUser, string sSchema, string sDataCategory, dyn_string &dsExInfo)
{
    string sql = "DELETE "
                 " FROM meta_category_mapping "
                 "WHERE user_name=:userName "
                 "  AND schema_name=:schemaName "
                 "  AND data_category_id=:dataCatId ";

    mapping params;   
    params[":userName"]  = sUser; 
    params[":schemaName"] = sSchema;      
    params[":dataCatId"]  = sDataCategory; 
    
    if (unLHCLoggingDBSql_executeWithParameterMapping(sql, params) == false) {
        fwException_raise(dsExInfo, "DB ERROR", "deleting data category mapping", "");
        return;
    }
    
    unLHCLoggingDBSql_commit(dsExInfo);
}


/**
* Execute SQL query to get session types
*
* @param    dsExInfo           output, Exception information
* @return   dsSessionTypes     output, list of session types
*/
public dyn_string unLHCLoggingDBSql_getSessionTypes(dyn_string &dsExInfo)
{
    string sql = "SELECT DISTINCT session_type "
                 "  FROM rdb2lhc.audit_session " +
                 "ORDER BY REGEXP_SUBSTR(session_type, '^\\D*') NULLS FIRST, ""TO_NUMBER(REGEXP_SUBSTR(session_type, '\\d+'))";  // to order numbers without leading zero right

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, true);

    dyn_string dsSessionTypes;
    if (dynlen(dsExInfo) > 0) { return dsSessionTypes; }
    if (dynlen(sqlResult) <= 0) { return dsSessionTypes; }
    dsSessionTypes = sqlResult[1];
    return dsSessionTypes;
}


/**
* Register new elements for logging
*
* @param dsElementId       input, list of element ids
* @param dsVariableName    input, list of variable names
* @param dsRDBSource       input, list of RDB sources
* @param dsUnit            input, list of units
* @param dsDescription     input, list of descriptions
* @param dsDataCategoryId  input, list of data category ids
* @param dsHierarchy       input, list of Hierarchies
* @param sUser             input, username from meta_users
* @param sPass             input, password from meta_users
*
* @return                  output, true: success, false: error
*/
public bool unLHCLoggingDBSql_registerDataForLogging(
        const dyn_string &dsElementId,
        const dyn_string &dsVariableName,
        const dyn_string &dsRDBSource,
        const dyn_string &dsUnit,
        const dyn_string &dsDescription,
        const dyn_string &dsDataCategoryId,
        const dyn_string &dsHierarchy,
        const string sUser,
        const string sPass)
{
    dyn_dyn_mixed ddmVariablesToInsert;
    dyn_dyn_mixed ddmHierarchiesToInsert;

    unLHCLoggingDBUtils_animateProgressBar(0);

    // Prepare Values to be inserted into the DB
    int numberOfElements = dynlen(dsElementId);
    for (int i = 1; i <= numberOfElements; i++) {

        // prepare meta_variables
        dynAppend(ddmVariablesToInsert, makeDynMixed(dsElementId[i],
                  dsVariableName[i],
                  dsRDBSource[i],
                  dsUnit[i],
                  dsDescription[i],
                  dsDataCategoryId[i]));
        // prepare meta_hierarchies
        if (dsHierarchy[i] != "") {
            dynAppend(ddmHierarchiesToInsert, makeDynMixed(dsElementId[i],
                      dsHierarchy[i]));
        }
    }

    unLHCLoggingDBUtils_animateProgressBar(10);

    // Insert into GTT_STAGE_VARIABLES
    if (fwDbExecuteSqlStatement(g_unLHCLoggingDBSql_dbConnection, "DELETE FROM rdb2lhc.gtt_stage_variables") != 0 ) {
        DebugTN("Error deleting gtt_stage_variables");
        return false;
    }
    string sql = "INSERT INTO rdb2lhc.gtt_stage_variables(ELEMENT_ID,VARIABLE_NAME,RDB_SOURCE,UNIT,DESCRIPTION,DATA_CATEGORY_ID) " +
                 "VALUES(:data1,:data2,:data3,:data4,:data5,:data6)";
    if (unLHCLoggingDBSql_executeWithBoundData(sql, ddmVariablesToInsert) == false) {
        DebugTN("Error inserting values into gtt_stage_variables");
        return false;
    }

    unLHCLoggingDBUtils_animateProgressBar(30);

    // Insert into GTT_STAGE_HIERARCHIES
    if (dynlen(ddmHierarchiesToInsert) > 0) {
        if (fwDbExecuteSqlStatement(g_unLHCLoggingDBSql_dbConnection, "DELETE FROM rdb2lhc.gtt_stage_hierarchies") != 0 ) {
            DebugTN("Error deleting gtt_stage_hierarchies");
            return false;
        }
        string sql = "INSERT INTO rdb2lhc.gtt_stage_hierarchies(ELEMENT_ID,HIERARCHY_PATH) " +
                     "VALUES(:data1,replace(:data2,'/','->'))";
        if (unLHCLoggingDBSql_executeWithBoundData(sql, ddmHierarchiesToInsert) == false) {
            DebugTN("Error inserting values into gtt_stage_hierarchies");
            return false;
        }
    }

    unLHCLoggingDBUtils_animateProgressBar(60);

    // Register metadata
    mapping params;
    params[":username"] = sUser;
    params[":password"] = sPass;
    string sql = "begin pvss_data_transfer_mgr.register_metadata(:username,:password); end;" ;
    if (unLHCLoggingDBSql_executeWithParameterMapping(sql, params) == false) {
        DebugTN("Registering Data Failed");
        return false;
    }

    unLHCLoggingDBUtils_animateProgressBar(100);

    return true;
}


/**
* Query current database system time
*
* @param dsExInfo     output, Exception information

* @return  sqlResult  output, database systimestamp as UTC
*/
public time unLHCLoggingDBSql_getDbSysTimeUtc(dyn_string &dsExInfo)
{
    string sql = "SELECT SYS_EXTRACT_UTC(SYSTIMESTAMP) FROM DUAL";
    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, false, false);

    time t = 0;
    if (dynlen(sqlResult) == 1) {
        t = sqlResult[1][1];
    }
    return t;
}


/**
* Query actual logging status for all registered elements of a specific rdb-source
*
* @param sDbSchema   input, DbSchema e.g.: RDBTEST01
* @param dsExInfo     output, Exception information

* @return  sqlResult  output, Result of the query (row wise)
*
* column 1: variable_name
* column 2: created_utc_stamp
* column 3: last_logged_utc_stamp
* column 4: no_recs_logged
* column 5: max_rows
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getLvsInfo(string sDbSchema, dyn_string &dsExInfo)
{
    dyn_dyn_mixed ddmResult;
    
    // Prequery rdb sources 
    dyn_string dsRdbSource = unLHCLoggingDBSql_getRdbSource(sDbSchema,dsExInfo);
    if (dynlen(dsExInfo)>0) return ddmResult; // return empty result
    
    string sql = "SELECT "
                 " mv.variable_name, "
                 " lvs.created_utc_stamp, "
                 " lvs.last_logged_utc_stamp, "
                 " lvs.no_recs_logged, "
                 " adtg.max_rows "
                 "FROM rdb2lhc.meta_variables mv "
                 " LEFT OUTER JOIN rdb2lhc.meta_variable_categories mvc "
                 "   ON mv.data_category_id = mvc.data_category_id "
                 " LEFT OUTER JOIN rdb2lhc.log_var_status_max lvsm "
                 "   ON mvc.data_transfer_group_id = lvsm.data_transfer_group_id "
                 " LEFT OUTER JOIN rdb2lhc.log_variable_status lvs "
                 "   ON mv.variable_id = lvs.variable_id "
                 "  AND mvc.data_transfer_group_id = lvs.data_transfer_group_id "
                 "  AND lvsm.created_utc_stamp = lvs.created_utc_stamp "
                 " LEFT OUTER JOIN rdb2lhc.admin_data_transfer_groups adtg "
                 "   ON mvc.data_transfer_group_id = adtg.data_transfer_group_id "
                 "WHERE mv.rdb_source = :RdbSource "
                 "  AND mv.obsolete_since_utc IS NULL";
       
    int dsRdbSourceLen = dynlen(dsRdbSource);
    for(int i=1; i<=dsRdbSourceLen; i++) { 

        mapping params;
        params[":RdbSource"] = dsRdbSource[i] ;

        dyn_dyn_mixed ddmTempResult = unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo, false, false);
        if (dynlen(dsExInfo)>0) return ddmResult; // return empty result
           
        int  ddmTempResultLen = dynlen(ddmTempResult);    
        for(int i=1; i<=ddmTempResultLen; i++) {
            dynAppend(ddmResult[i], ddmTempResult[i]); 
        }
    }
    
    return ddmResult;              
}


/**
* Get the 3 most recent entries in log variable status table
* within the last hour for a specified variable
*
* This function is used by the test datapoints of system integrity plugin
* but it can be used for any variable.
*
* @param sVarName       input, Variable Name of the test datapoint
* @param dsExInfo       output, Exception information

* @return  sqlResult    output, Result of the query (row wise)
*
* sqlResult[1][1]: lvs.created_utc_stamp (most recent)
* sqlResult[2][1]: lvs.created_utc_stamp (2nd most recent)
* sqlResult[3][1]: lvs.created_utc_stamp (3rd most recent)
* sqlResult[1][2]: lvs.last_logged_utc_stamp (most recent)
* sqlResult[2][2]: lvs.last_logged_utc_stamp (2nd most recent)
* sqlResult[3][2]: lvs.last_logged_utc_stamp (3rd most recent)
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getLvsForTestDpe(string sVarName, dyn_string &dsExInfo)
{
    string sql = "SELECT * FROM ( " +
                 "  SELECT lvs.created_utc_stamp " +
                 "       , lvs.last_logged_utc_stamp " +
                 "       , lvs.last_numeric_value " +
                 "   FROM rdb2lhc.meta_variables mv JOIN rdb2lhc.log_variable_status lvs " +
                 "     ON mv.variable_id = lvs.variable_id " +
                 "  WHERE lvs.created_utc_stamp BETWEEN (SYS_EXTRACT_UTC(SYSTIMESTAMP)-1/24) " +
                 "    AND SYS_EXTRACT_UTC(SYSTIMESTAMP) " +
                 "    AND mv.variable_name=UPPER(:VarName) " +
                 "  ORDER BY lvs.created_utc_stamp DESC " +
                 ") " +
                 "WHERE ROWNUM <=3 ";

    mapping params;
    params[":VarName"] = sVarName;

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo, false, false);
}


/**
* Query for transfer statistics per transfer job
*
* @param job           input, name of the transfer job
* @param tStart        input, utc start time of time window
* @param tEnd          input, utc end time of time window
* @param bUTC          input, true=show timestamps as UTC, false=show timestamps as localtime
* @param dsExInfo      output, Exception information

* @return sqlResult    output, Result of the query (column wise)
*
* line 1: session_id
* line 2: session_start_utc
* line 3: session_type
* line 4: time_in_s
* line 5: no_of_recs
* line 6: time_in_s
* line 7: no_of_recs
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getTransferStats(string job, time tStart, time tEnd, bool bUTC, dyn_string &exceptionInfo)
{
    string sql;
    sql = "select ao1.session_id, " +
          ( bUTC ? " s.session_start_utc, " : " from_tz(s.session_start_utc, 'UTC') at local, " ) +
          " s.session_type, " +
          " ao1.time_in_s t1, " +
          " ao1.no_of_recs/1000 nrec1k, " +  // get no of records in k
          " ao2.time_in_s t2, " +
          " ao2.no_of_recs/1000 nrec2k " +   // get no of records in k
          "from audit_operation ao1 " +
          " join audit_operation ao2 on ao1.session_id=ao2.session_id" +
          " left outer join audit_session s on s.session_id=ao1.session_id " +
          " where ao1.op_type like 'COLLECT_%' and ao2.op_type like 'LOG%' " +
          " and session_type=:transferid " +
          "AND SESSION_START_UTC BETWEEN :t1 and :t2 " +
          " order by session_id desc ";

    mapping params;
    params[":transferid"] = job;
    params[":t1"] = tStart;
    params[":t2"] = tEnd;

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, exceptionInfo, true);
}


/**
* Execute SQL query to get transfer job IDs
*
* @param   dsExInfo           output, Exception information
* @return  sqlResult          output, result of the query (column wise)
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getTransferJobIds(dyn_string &dsExInfo)
{
    string sql = "SELECT data_transfer_group_id "
                 "FROM rdb2lhc.admin_data_transfer_groups "
                 "  ORDER BY data_transfer_group_id ASC ";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, true);
}


/**
* Execute SQL query to get information of RDB transfer jobs
*
* @param   sRdbSource         input, show transfer jobs related to this RDB SOURCE
* @param   bUTC               input, true=show timestamps as UTC, false=show timestamps as localtime
* @param   dsExInfo           output, Exception information

* @return  sqlResult          output, Result of the query (column wise)
*
* line 1: job name
* line 2: job enabled
* line 3: job state
* line 4: last start date
* line 5: last run duration
* line 6: next run date
* line 7: load in percent
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getTransferJobDetails(string sRdbSource, bool bUTC, dyn_string &dsExInfo)
{
    string sql;
    sql = "SELECT " +
          "  job_name, " +
          "  enabled, " +
          "  state, " +
          "  TO_CHAR(" + ( bUTC ? "SYS_EXTRACT_UTC(last_start_date)" : "last_start_date" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
          "  3600*EXTRACT(hour FROM last_run_duration) + 60*EXTRACT(minute from last_run_duration) + EXTRACT(second from last_run_duration), " +
          "  TO_CHAR(" + ( bUTC ? "SYS_EXTRACT_UTC(next_run_date)" : "next_run_date" ) + ",'YYYY/MM/DD HH24:MI:SS'), "        +
          "  (3600*EXTRACT(hour FROM last_run_duration) + 60*EXTRACT(minute from last_run_duration) + EXTRACT(second from last_run_duration)) / 3.0 " +
          "FROM " +
          "  dba_scheduler_jobs " +
          "WHERE " +
          "  owner='RDB2LHC' " +
          "  AND job_action LIKE 'ldb_data_transfer_service.transfer_data_to_ldb%' " +
          "  AND job_action IN ( " +
          "    SELECT DISTINCT 'ldb_data_transfer_service.transfer_data_to_ldb(' || mvc.data_transfer_group_id || ');' " +
          "    FROM rdb2lhc.meta_variables mv "
          "      LEFT JOIN rdb2lhc.meta_variable_categories mvc ON mv.data_category_id = mvc.data_category_id " +
          ( sRdbSource ? "    WHERE rdb_source = '" + sRdbSource + "' " : "" ) +
          "  ) " +
          "ORDER BY REGEXP_SUBSTR(job_name, '^\\D*') NULLS FIRST, TO_NUMBER(REGEXP_SUBSTR(job_name, '\\d+')) ";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, true);
}


/**
* SQL query to get current status of transfer jobs
*
* @param   sRdbSource         input, show transfer jobs related to this RDB SOURCE
* @param   dsExInfo           output, Exception information

* @return  sqlResult          output, Result of the query
*
* sqlResult[1]: job name
* sqlResult[2]: job enabled
* sqlResult[3]: job state
* sqlResult[4]: last start date (utc)
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getTransferJobStatus(string sRdbSource, dyn_string &dsExInfo)
{
    string sql;
    sql = "SELECT "
          "  job_name, "
          "  enabled, "
          "  state, "
          "  SYS_EXTRACT_UTC(last_start_date) "
          "FROM "
          "  dba_scheduler_jobs "
          "WHERE "
          "  owner='RDB2LHC' "
          "  AND job_action LIKE 'ldb_data_transfer_service.transfer_data_to_ldb%' "
          "  AND job_action IN ( "
          "    SELECT DISTINCT 'ldb_data_transfer_service.transfer_data_to_ldb(' || mvc.data_transfer_group_id || ');' "
          "    FROM rdb2lhc.meta_variables mv 	"
          "      LEFT JOIN rdb2lhc.meta_variable_categories mvc ON mv.data_category_id = mvc.data_category_id "
          "    WHERE rdb_source LIKE :RdbSource "
          "  ) "
          "ORDER BY REGEXP_SUBSTR(job_name, '^\\D*') NULLS FIRST, TO_NUMBER(REGEXP_SUBSTR(job_name, '\\d+')) ";

    mapping params;
    params[":RdbSource"] = sRdbSource;

    return unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo, false, false);
}


/**
* Execute SQL query to get information about the latest session
*
* @param   sJobNumber        input, job number as a string, e.g.: "7"
* @param   dsExInfo          output, Exception information
* @param   bUTC              input, true=show timestamps as UTC, false=show timestamps as localtime

* @return sqlResult

column 1: session id
column 2: session type
column 3: session start
column 4: session end
column 5: operation type
column 6: operation time
column 7: operation no of records

*/
public dyn_dyn_mixed unLHCLoggingDBSql_getSessionInfo( const string sJobNumber, dyn_string &dsExInfo, bool bUTC = false)
{
    string sql;
    sql = "SELECT " +
          "  as_.session_id, " +
          "  as_.session_type, " +
          "  TO_CHAR(" + ( bUTC ? "as_.session_start_utc" : "FROM_TZ(as_.session_start_utc, 'UTC') AT LOCAL" ) + ", 'YYYY/MM/DD HH24:MI:SS'), " +
          "  TO_CHAR(" + ( bUTC ? "as_.session_end_utc" : "FROM_TZ(as_.session_end_utc, 'UTC') AT LOCAL" ) + ",'YYYY/MM/DD HH24:MI:SS'), " +
          "  ao.op_type, " +
          "  ao.time_in_s, " +
          "  ao.no_of_recs " +
          "FROM ( " +
          "  SELECT * "
          "  FROM ( " +
          "    SELECT * "
          "    FROM rdb2lhc.audit_session "
          "    WHERE session_type='LDB_DATA_TRANSFER_" + sJobNumber + "' "
          "      ORDER BY session_id desc) "
          "  WHERE rownum <=1) as_ " +
          "    LEFT OUTER JOIN rdb2lhc.audit_operation ao ON as_.session_id=ao.session_id " +
          "ORDER BY ao.op_type ASC" ;  // first collect then log

    return unLHCLoggingDBSql_queryData(sql, dsExInfo);
}


/**
* Get all data categories related to a DB-schema
*
* @param  sDbSchema        input, list of user names
* @param  dsExInfo         output, Exception information
*
* @return dsDataCategories output, list of data categories
*/
public dyn_string unLHCLoggingDBSql_getDataCategoriesOfSchema(string sDbSchema, dyn_string &dsExInfo)
{
    string sql = "SELECT DISTINCT ( mvc.data_category_name || '/' || mvc.data_category_sub_name ) " +
                 " FROM rdb2lhc.meta_variables mv " +
                 " INNER JOIN rdb2lhc.meta_variable_categories mvc ON mv.data_category_id = mvc.data_category_id " +
                 "WHERE mv.rdb_source LIKE '" + sDbSchema + ".%' AND mv.obsolete_since_utc IS NULL";

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, true);

    dyn_string dsDataCategories;
    if (dynlen(sqlResult) > 0) {
        dsDataCategories = sqlResult[1];
    }
    return dsDataCategories;
}


/**
* Get data category id of a data category
*
* @param sDataCategory  input, "data_category_name/data_category_sub_name"
* @param dsExInfo       output, Exception information
*
* @return               output, data category id as a string
*/
public string unLHCLoggingDBSql_getDataCategoryId(string sDataCategory, dyn_string &dsExInfo)
{
    string sql = "SELECT data_category_id " +
                 " FROM rdb2lhc.meta_variable_categories " +
                 "WHERE (data_category_name || '/' || data_category_sub_name) = :cat ";

    mapping params;
    params[":cat"] = sDataCategory;

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo);

    if (dynlen(sqlResult) == 1) {
        return (sqlResult[1][1]);
    }

    return ("");
}


/**
* Unregister single variable from logging
*
* @param sVariableName    input, variable name
* @return                 output, true: success, false: error
*/
public bool unLHCLoggingDBSql_unregisterVariableFromLogging(const string sVariableName)
{
    mapping params;
    params[":varName"] = sVariableName;

    string sql = "BEGIN " +
                 "  pvss_data_transfer_mgr.mark_var_obsolete(UPPER(:varName)); " +
                 "END;" ;

    return unLHCLoggingDBSql_executeWithParameterMapping(sql, params);
}


/**
* Unregister dB Schema from logging
*
* @param sDbSchema       input, Name of the db schema
* @return                output, true: success, false: error
*/
public bool unLHCLoggingDBSql_unregisterSchemaFromLogging(const string sDbSchema)
{
    dyn_string dsExInfo;

    string sql = "SELECT DISTINCT rdb_source " +
                 " FROM rdb2lhc.meta_variables " +
                 "WHERE rdb_source LIKE '" + sDbSchema + ".%'";

    dyn_dyn_mixed ddmRdbSource = unLHCLoggingDBSql_queryData(sql, dsExInfo);

    if (dynlen(dsExInfo) > 0) {return false;};

    int iDdmRdbSourceLen = dynlen(ddmRdbSource);
    for (int i = 1; i <= iDdmRdbSourceLen; i++) {
        mapping params;
        params[":RdbSource"] = ddmRdbSource[i][1];

        string sql = "BEGIN " +
                     "  pvss_data_transfer_mgr.mark_schema_obsolete(UPPER(:RdbSource)); " +
                     "END;" ;

        if (unLHCLoggingDBSql_executeWithParameterMapping(sql, params) == false) {
            return false;
        }
    }

    return true;
}


/**
* Find inconsistency between registered data (meta_variables) and dpes (elements)
*
* @param sDbSchema    input, name of the db schema to check
* @param dsExInfo     output, Exception information
*
* @return             output, list of inconsistent variables
*
* column 1: variable_id
* column2: variable_name
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getInconsistentData(const string sDbSchema, dyn_string &dsExInfo)
{
    string sql = "SELECT " +
                 "  mv.variable_id, " +
                 "  mv.variable_name " +
                 "FROM rdb2lhc.meta_variables mv " +
                 "  LEFT JOIN " + sDbSchema + ".elements el ON mv.element_id = el.element_id " +
                 "WHERE mv.rdb_source LIKE '" + sDbSchema + ".%' " +
                 "  AND mv.obsolete_since_utc IS NULL " +
                 "  AND el.element_id IS NULL ";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, false, false);
}


/**
* Check wheter an element is registered and not obsolete
*
* @param sVariableName  input, variable name
* @param dsExInfo       output, Exception information
*
* @return               output, true: ok, false: problem
*/
public bool unLHCLoggingDBSql_isElementRegistered(const string sVariableName, dyn_string &dsExInfo)
{
    string sql = "SELECT count(1) "
                 "  FROM rdb2lhc.meta_variables " +
                 " WHERE variable_name=UPPER(:varName) " +
                 "   AND logging_required = 'Y'" +
                 "   AND obsolete_since_utc IS NULL";

    mapping params;
    params[":varName"] = sVariableName;

    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryDataWithParameterMapping(sql, params, dsExInfo, false, false);

    if (dynlen(dsExInfo) > 0) {return false;};
    if (dynlen(sqlResult) <= 0) {return false;};
    return (sqlResult[1][1] > 0);
}


/**
* Get all elements in a specific db schema
*
* @param sDbSchema      input, db schema to look for elements
*
* @return               output, true: ok, false: problem
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getElements(const string sDbSchema, dyn_string &dsExInfo)
{
    string sql = "SELECT element_name, element_id " +
                 "FROM " + sDbSchema + ".elements " + 
                 "WHERE EVENT=1";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, true, false);
}


/**
* Get all elements in a specific db schema
*
* @param sDbSchema      input, db schema to look for elements
*
* @return               output, true: ok, false: problem
*/
public dyn_dyn_mixed unLHCLoggingDBSql_getElementsExceptInternalDPs(const string sDbSchema, dyn_string &dsExInfo)
{    
    string sql ="SELECT " +  
                "  e.element_name " 
                "FROM " + sDbSchema + ".elements e " 
                "INNER JOIN " + sDbSchema + ".DPT t " 
                "  ON e.dpt_id = t.dpt_id " 
                "WHERE t.dptname NOT LIKE \'\\_%\' ESCAPE \'\\\'" 
                "  AND e.event = 1 ";

    return unLHCLoggingDBSql_queryData(sql, dsExInfo, true, false);
}


/**
* Trigger retransfer for given RDB_SOURCE starting from a certain START_TIME
*
* @param sDbSchema      input, RDB_SOURCE to retransfer
* @param tStart         input, START_TIME for retransfer (localtime)
*
* @return               output, true:sucess, false:error
*/
public bool unLHCLoggingDBSql_triggerRetransferForRdbSource(string sRdbSource, time tStart)
{    
    mapping params;
    params[":RdbSource"] = sRdbSource;
    params[":StartTime"] = tStart;
      
    string sql = "begin " +
                 "pvss_data_transfer_mgr.reset_last_checked_for_schema(:RdbSource,SYS_EXTRACT_UTC(:StartTime)); " +
                 "end;" ;
    
    return unLHCLoggingDBSql_executeWithParameterMapping(sql, params);
}


/**
* Get all RDB_SOURCES, that are currently registered
*
* @param dsExInfo       output, Exception information
*
* @return               output, List of RDB_SOURCES
*/
public dyn_string unLHCLoggingDBSql_getAllRdbSources(dyn_string &dsExInfo)
{
    string sql = "SELECT DISTINCT rdb_source "
                 "FROM meta_variables "
                 "WHERE obsolete_since_utc IS NULL "
                 "ORDER BY rdb_source ASC";    
    
    dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo, true);

    if (dynlen(dsExInfo) > 0) {
        return makeDynString();
    }

    if (dynlen(sqlResult) <= 0) {
        fwException_raise(dsExInfo, "DB ERROR", "No RDB_SOURCE found", "");
        return makeDynString();
    }

    return sqlResult[1];
}



//@}
